import { afterEach, beforeEach, describe, expect, test, vi } from 'vitest';
import {
  AgentType,
  CMType,
  HMType,
  type HostMessage,
  type Transaction,
  type TransactionOptions,
  type Withdrawal,
} from '../../types';
import EmbeddedComponent from '../EmbeddedComponent';

const url = 'https://example.com';
const parentId = 'parent-container';
const value: Partial<Transaction> = { amountDecimal: 100 };

describe('EmbeddedComponent', () => {
  describe('creation', () => {
    let component: EmbeddedComponent<Withdrawal, TransactionOptions>;
    beforeEach(() => {
      component = new EmbeddedComponent<Withdrawal, TransactionOptions>(url, {
        amountDecimal: 100,
      });
    });

    test('sets value', () => {
      expect(component.value).toStrictEqual(value);
    });

    test('sets url', () => {
      expect(component.url).toBe(url);
    });
  });

  describe('iframe', () => {
    let document: Document;
    let parentElement: HTMLElement;
    let iframe: HTMLIFrameElement;
    let window: Window;
    let iframeWindow: Window;

    beforeEach(() => {
      parentElement = {
        appendChild: vi.fn(),
      } as unknown as HTMLElement;

      iframe = {
        style: {},
        src: '',
        contentWindow: iframeWindow,
        setAttribute: vi.fn(),
        addEventListener: vi.fn(),
      } as unknown as HTMLIFrameElement;

      document = {
        createElement: vi.fn().mockReturnValue(iframe),
        querySelector: vi.fn().mockReturnValue(parentElement),
      } as unknown as Document;

      window = {
        addEventListener: vi.fn(),
      } as unknown as Window;

      vi.stubGlobal('document', document);
      vi.stubGlobal('window', window);
    });

    afterEach(() => {
      vi.unstubAllGlobals();
    });

    test('mount', () => {
      const component = new EmbeddedComponent<Withdrawal, TransactionOptions>(
        url,
        value,
      );
      component.mount(parentId);
      expect(iframe?.src).toBe(url + '&embedded=true');
      expect(parentElement.appendChild).toHaveBeenCalledWith(iframe);
    });

    test('on', () => {
      const component = new EmbeddedComponent<Withdrawal, TransactionOptions>(
        url,
        value,
      );
      const messageType = 'customEvent';
      const callback = vi.fn();
      component.on(messageType, callback);
      component.mount(parentId);

      // Simulate receiving a message of the specified type
      // and assert that the callback is called with the correct payload.
    });

    test('off', () => {
      const component = new EmbeddedComponent<Withdrawal, TransactionOptions>(
        url,
        value,
      );
      const messageType = 'customEvent';
      const callback = vi.fn();
      component.on(messageType, callback);

      // Simulate receiving a message of the specified type
      // and assert that the callback is called with the correct payload.

      component.off(messageType, callback);

      // Simulate receiving another message of the specified type
      // and assert that the callback is not called.
    });

    describe('connected to iframe', () => {
      let messageHandler: (event: MessageEvent) => void;
      let port: MessagePort;
      let handshake: MessageEvent;
      let component: EmbeddedComponent<Withdrawal, TransactionOptions>;
      let options: TransactionOptions;

      beforeEach(() => {
        port = {
          postMessage: vi.fn(),
          start: vi.fn(),
          onmessage: null,
        } as unknown as MessagePort;
        handshake = {
          source: iframe.contentWindow,
          ports: [port],
        } as unknown as MessageEvent;
        window.addEventListener = vi
          .fn()
          .mockImplementation((event, handler) => {
            messageHandler = handler;
          });
        options = { allowedAgentTypes: [AgentType.PRIVATE, AgentType.VASP] };
        component = new EmbeddedComponent<Withdrawal, TransactionOptions>(
          url,
          value,
          options,
        );
        component.mount(parentId);

        expect(messageHandler).toBeDefined();
        messageHandler(handshake);
      });

      test('send', () => {
        const message: HostMessage<Withdrawal, TransactionOptions> = {
          type: HMType.UPDATE,
          value: {
            requestId: 'id',
          },
          options: { allowedAgentTypes: [AgentType.PRIVATE] },
        };
        component.send(message);
        expect(port.postMessage).toHaveBeenCalledWith(message);
      });

      describe('update', () => {
        test('with out options', () => {
          const newValue: Partial<Transaction> = { amountDecimal: 200 };

          const message: HostMessage<Withdrawal, TransactionOptions> = {
            type: HMType.UPDATE,
            value: newValue,
            options,
          };

          component.update(newValue);

          expect(component.value).toBe(newValue);
          expect(component.options).toBe(options);
          expect(port.postMessage).toHaveBeenCalledWith(message);
        });

        test('with options', () => {
          const newValue: Partial<Transaction> = { amountDecimal: 200 };
          const newOptions = { allowedAgentTypes: [AgentType.PRIVATE] };

          const message: HostMessage<Withdrawal, TransactionOptions> = {
            type: HMType.UPDATE,
            value: newValue,
            options: newOptions,
          };

          component.update(newValue, newOptions);

          expect(component.value).toBe(newValue);
          expect(component.options).toBe(newOptions);
          expect(port.postMessage).toHaveBeenCalledWith(message);
        });
      });
      // TODO not sure this actually does what I want it to do
      test('completion', async () => {
        const response = { valid: true, value: 'completed' };
        expect(port.onmessage).toBeDefined();

        const promise = component.completion();
        // @ts-expect-error - we've already verified that port exists
        port.onmessage({
          data: { type: CMType.COMPLETE, response },
        });

        promise.then((result) => {
          expect(result).toBe(response);
        });
        return promise;
        // Assert that the completion promise resolves with the expected response.
      });
    });
  });
});
